#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# 增/删/改/查/获取所有

# POST: {srv-name}/models
# DELETE: {srv-name}/models/{instID}
# POST: {srv-name}/models/{instID}
# GET: {srv-name}/models/{instID}
# GET: {srv-name}/models

import json
from pdb import set_trace as stop

from tornado.escape import json_decode
from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from tornado.web import Application, RequestHandler

from .sql import Entry


class CustomHandler(RequestHandler):

    def get_argument(self, id: str):
        if self.request.headers['Content-Type'] == 'application/json':
            args = json_decode(self.request.body)
            return args[id]
        return super(CustomHandler, self).get_argument(id)


# # SQLITE
# class Entry(Base):
#     __tablename__ = 'model'
#     id = Column(String, primary_key=True, index=True, unique=True)
#     content = Column(String)
#     def __repr__(self):
#         return json.dumps({
#             "id": self.id,
#             "content": self.content
#         })

class Http():
    def __init__(self, config: dict, srv_name: str, sql=None) -> None:
        self.config = config
        self.srv_name = srv_name
        self.sql = sql
        self.srv = None

        # print('sqlite:///{0}/dm.db'.format(self.config["WORKDIR"]))
        # engine = create_engine('sqlite:///{0}/dm.db'.format(self.config["WORKDIR"]))
        # Entry.__table__.create(engine, checkfirst=True)
        # self.sql = sessionmaker(bind=engine)()

    def Build(self):
        self.appList = [
            ('/{0}/models'.format(self.srv_name), ModelProc, dict(sql=self.sql)),
            ('/{0}/models/(\w+)'.format(self.srv_name),
             ModelProcID, dict(sql=self.sql)),
            ('/{0}/models_exist/(\w+)'.format(self.srv_name),
             ModelExist, dict(sql=self.sql)),
            ('/{0}/property/(\w+)'.format(self.srv_name),
             Property, dict(srv=self.srv)),
            ('/{0}/action/(\w+)'.format(self.srv_name),
             Action, dict(srv=self.srv)),
            ('/{0}/event/(\w+)'.format(self.srv_name), Event, dict(srv=self.srv)),
            ('/{0}/mqtt/(\w+)'.format(self.srv_name), MqttOp, dict(srv=self.srv)),
            ('/{0}/devices'.format(self.srv_name),
             DevProc, dict(sql=self.sql, srv=self.srv)),
            ('/{0}/devices/(\w+)'.format(self.srv_name),
             DevProcID, dict(sql=self.sql, srv=self.srv)),
        ]
        self.httpApp = Application(self.appList)
        self.httpServer = HTTPServer(self.httpApp)
        self.httpServer.bind(self.config['PORT'])
        if 'HTTP_THREAD' in self.config:
            self.httpServer.start(self.config['HTTP_THREAD'])
        else:
            self.httpServer.start(1)

    def Run(self):
        IOLoop.current().start()


class ModelProc(CustomHandler):
    def initialize(self, sql):
        self.sql = sql

    # Add new model
    def post(self):
        print(self.request.headers['Content-Type'])
        model = self.sql.query(Entry).filter_by(
            id=self.get_argument('id')).first()

        if model == None:
            # Add
            model = Entry(id=self.get_argument('id'),
                          content=self.get_argument('content'))
            print("Entry Added")
            self.sql.add(model)
        else:
            # Update
            model.content = self.get_argument('content')
            print("Entry exists, update")
        self.sql.commit()
        self.set_status(200)
        self.write('{0}'.format(model))

    # Get all models
    def get(self):
        print("Get Models")
        model = self.sql.query(Entry).all()
        self.set_status(200)
        self.write("{0}".format(model))


class ModelProcID(CustomHandler):
    def initialize(self, sql):
        self.sql = sql

    # Delete model
    def delete(self, id):
        print("Delete Entry - ID: " + id)
        model = self.sql.query(Entry).filter_by(id=id).first()
        self.sql.delete(model)
        self.sql.commit()
        self.set_status(200)
        self.write("{0}".format(model))

    # Update model
    def post(self, id):
        print("Update Entry - ID: " + id)
        model = self.sql.query(Entry).filter_by(id=id).first()
        model.content = self.get_argument('content')
        self.sql.commit()
        self.set_status(200)
        self.write("{0}".format(model))

    # Get model
    def get(self, id: str):
        print("Get Entry - ID: " + id)
        model = self.sql.query(Entry).filter_by(id=id).first()
        self.set_status(200)
        self.write("{0}".format(model))


class ModelExist(CustomHandler):
    def initialize(self, sql):
        self.sql = sql

    # Check exist
    def get(self, id: str):
        model = self.sql.query(Entry).filter_by(id=id).first()
        self.set_status(200)
        if model == None:
            print("Entry Not Exist - ID: " + id)
            self.write("false")
        else:
            print("Entry Exist - ID: " + id)
            self.write("true")


# 读取/操作

# GET: {srv-name}/property/{property-name}
# POST: {srv-name}/action/{function-name}
# POST & MQTT: {srv-name}/event/{event-name}

class Property(CustomHandler):
    def initialize(self, srv):
        self.srv = srv

    def post(self, key: str):
        content = self.get_argument('content')
        is_read = self.get_argument('is_read')
        if is_read:
            self.write(json.dumps(self.srv.ReadProperty(
                key, content), ensure_ascii=False))
        else:
            self.write(json.dumps(self.srv.WriteProperty(
                key, content), ensure_ascii=False))


class Action(CustomHandler):
    def initialize(self, srv):
        self.srv = srv

    def get(self, func_name: str):
        self.write(json.dumps(self.srv.Execute(func_name), ensure_ascii=False))

    def post(self, func_name: str):
        content = self.get_argument('content')
        self.write(json.dumps(self.srv.Execute(
            func_name, content), ensure_ascii=False))


class Event(CustomHandler):
    def initialize(self, srv):
        self.srv = srv
        self.mqtt = srv.mqtt

    def post(self, event_name: str):
        setup = int(self.get_argument('setup'))  # 0 = off, 1 = on
        pub_topic = self.get_argument('pub_topic')
        qos = 1
        try:
            qos = int(self.get_argument('qos'))
        except:
            pass

        print("Event Subscribed: {0}, {1}".format(setup, pub_topic))

        def on_event_callback(content):
            if (setup > 0):
                print("Event auto-pubed: {0}, {1}".format(setup, pub_topic))
                self.mqtt.publish(pub_topic, json.dumps(
                    {'content': content}, ensure_ascii=False), qos)

        if not hasattr(self.srv.app, 'ecbs'):
            print("ecb not initiated")
            self.write("Error: ecb not initiated")
        else:
            self.srv.app.ecbs[event_name] = on_event_callback
            self.write("OK")

# MQTT 操作 (Flow)


class MqttOp(CustomHandler):
    def initialize(self, srv):
        self.srv = srv
        self.mqtt = srv.mqtt

    def post(self, func_name: str):
        op = self.get_argument('op')
        sub_topic = None

        try:
            sub_topic = self.get_argument('sub_topic')
        except:
            pass

        pub_topic = self.get_argument('pub_topic')
        qos = 1

        try:
            qos = int(self.get_argument('qos'))
        except:
            pass

        print("MQTT Op: {0}, {1}, {2}".format(op, sub_topic, pub_topic))

        # Automate: sub func input, pub func output
        if (op == "add"):
            if (sub_topic != None):
                self.mqtt.subscribe(sub_topic, qos)

            # Inject constants
            constants = {}
            try:
                constants = json.loads(self.get_argument('constants'))
            except:
                pass

            def on_subs_callback(client, userdata, msg):
                msg.payload = json.loads(msg.payload)
                msg.payload.update(constants)

                result = self.srv.Execute(func_name, msg.payload)
                if result is not None:
                    # Decode input
                    self.mqtt.publish(pub_topic, json.dumps(
                        {'content': self.srv.Execute(func_name, msg.payload)}, ensure_ascii=False), qos)
            # Sub input
            self.mqtt.message_callback_add(sub_topic, on_subs_callback)
        elif (op == "remove"):
            self.mqtt.unsubscribe(sub_topic)
            self.mqtt.message_callback_remove(sub_topic)
        self.write("OK")

# Protocol 协议服务专用


class DevProc(CustomHandler):
    def initialize(self, sql, srv):
        self.sql = sql
        self.srv = srv

    # Add new model
    def post(self):
        print(self.request.headers['Content-Type'])

        req = json.loads(self.request.body)
        req_json = json.dumps(req, ensure_ascii=False)
        id = req['id']

        model = self.sql.query(Entry).filter_by(id=id).first()

        if model == None:
            # Add
            model = Entry(id=id, content=req_json)
            print("Device Added")
            res = self.srv.AddDev(req)
            # if res is not None:
            #     self.write(str(res))
            #     return
            self.sql.add(model)
        else:
            # Update
            res = self.srv.UpdateDev(req)
            # if res is not None:
            #     self.write(str(res))
            #     return
            model.content = req_json
            print("Device exists, update")
        self.sql.commit()
        self.set_status(200)
        self.write('Device Added: {0}'.format(id))

    # Get all models
    def get(self):
        models = self.sql.query(Entry).all()
        for m in models:
            m.content = json.dumps(json.loads(m.content), ensure_ascii=False)
        models = json.dumps([{"id": m.id, "content": m.content}
                            for m in models], ensure_ascii=False)
        self.set_status(200)
        self.write(models)


class DevProcID(CustomHandler):
    def initialize(self, sql, srv):
        self.sql = sql
        self.srv = srv

    # Delete model
    def delete(self, id):
        print("Delete Device - ID: " + id)
        model = self.sql.query(Entry).filter_by(id=id).first()
        if model is None:
            self.write('Device does not exist - ID: ' + id)

        res = self.srv.DeleteDev(id)
        if res is not None:
            self.write(str(res))
            return

        self.sql.delete(model)
        self.sql.commit()
        self.set_status(200)
        self.write('Device Deleted: {0}'.format(id))
        if res == "exit":
            self.set_status(200)
            self.write('REBOOT, Device Deleted: {0}'.format(id))
            import os
            os._exit(1)

    # Get model
    def get(self, id: str):
        print("Get Device - ID: " + id)
        m = self.sql.query(Entry).filter_by(id=id).first()
        model = json.dumps(
            {"id": m.id, "content": m.content}, ensure_ascii=False)
        self.set_status(200)
        self.write(model)



import json
from pdb import set_trace as stop

from sqlalchemy import Integer, create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql.expression import true

from .config import Config
from .http import Http
from .mqtt import Mqtt
from .sql import Entry, Flow


class Service():
    default_config = {
        "ADDR": "127.0.0.1",
        "PORT": "10100",
        "WORKDIR": ".",
        "MQTT_ADDR": "localhost",
        "MQTT_PORT": 1883
    }

    def __init__(self, cfg: object, srv_name: str):
        self.config = Config(self.default_config)
        self.config.from_object(cfg)
        self.srv_name = srv_name

        # print('sqlite:///{0}/dm.db?charset=utf8'.format(self.config["WORKDIR"]))
        engine = create_engine(
            'sqlite:///{0}/dm.db'.format(self.config["WORKDIR"]))
        Entry.__table__.create(engine, checkfirst=True)
        Flow.__table__.create(engine, checkfirst=True)
        self.sql = sessionmaker(bind=engine)()

        self.http = Http(self.config, self.srv_name, sql=self.sql)
        self.http.srv = self

        self.mqtt = Mqtt(self.config['MQTT_ADDR'], self.config['MQTT_PORT'])
        self.mqtt.srv = self

        self.devmodels = {}
        self.gcbs = {}  # global callback set, 叠加回调针对单输入多输出的情况

    def Bind(self, app):
        self.app = app
        self.app.srv = self
        self.devs = self.app.devs
        self.http.Build()

    def AddDev(self, model):
        try:
            return self.app.AddDev(model)
        except Exception as e:
            return e

    def UpdateDev(self, model):
        try:
            return self.app.UpdateDev(model)
        except Exception as e:
            return e

    def DeleteDev(self, id):
        try:
            return self.app.DeleteDev(id)
        except Exception as e:
            return e

    def InitDev(self, model):
        try:
            return self.app.InitDev(model)
        except Exception as e:
            print(e)
            return str(e)

    def InitAllDevs(self):
        models = self.sql.query(Entry).all()
        flows = self.sql.query(Flow).all()
        try:
            for m in models:
                model = json.loads(m.content)
                self.InitDev(model)
                if len(flows) > 0:
                    for fs in flows:
                        flowModels = json.loads(fs.content)
                        for f in flowModels:
                            if f['dev_id'] == m.id:
                                self.devs[f['dev_id']].flow(
                                    'add', srv_name=f['srv_name'], sub_topic=f['sub_topic'], pub_topic=f['pub_topic'], qos=f['qos'])
        except Exception as e:
            print("[InitAllDevs error]:" + e)
            pass

    # def parse(self, model_str: str) -> object:

    def ReadProperty(self, key: str, content):
        print("Read property: " + key)
        try:  # Service mode
            return getattr(self.app, key)
        except:  # Protocol mode (need implementation)
            return self.app.ReadProperty(key, content)

    def WriteProperty(self, key: str, content):
        print("Write property: " + key)
        print(content)
        try:  # Service
            setattr(self.app, key, content)
        except:  # Protocol
            return self.app.WriteProperty(key, content)

    def Execute(self, func_name: str, content=None):
        print("Execute function: " + func_name)
        if (content == None):
            return getattr(self.app, func_name)()
        else:
            print(content)
            if isinstance(content, str):
                return getattr(self.app, func_name)(**(json.loads(content)))
            elif isinstance(content, dict):
                return getattr(self.app, func_name)(**content)
            else:
                return "Content type is not supported"



#!/usr/bin/env python
# --*--coding=utf-8--*--

from pdb import set_trace as stop
from metathing.metathing import MetaThing as MT
from metathing.http import Entry
from metathing.device import Device
import threading
import time
import sys
import json

class Dev(Device):
    def __init__(self, srv, model) -> None:
        super().__init__(srv, model)
        self.mqtt = self.srv.mqtt
        self.resources = {}

class Bluetooth():
    def __init__(self):
        self.ecbs = {}
        self.srv = None
        self.devs = {}
        self.deviceModel = {}

    def AddDev(self, model):
        dev = Dev(self.srv, model)
        self.devs[model['id']] = dev
        print("Model added: {0}".format(model['id']))
        return dev
    
    def UpdateDev(self, model):
        self.AddDev(model)
    
    def DeleteDev(self, id):
        del self.devs[id]
        quit()
        # TODO: Release?
    
    def InitDev(self, model):
        return self.AddDev(model)

    # Run after binding srv
    def Initialize(self):
        return

    def Ping(self):
        return "OK"


if __name__ == "__main__":
    filepath = sys.path[0]

    config = {
        "ADDR": "0.0.0.0",
        "PORT": 16543,
        "WORKDIR": '.',
        "MQTT_ADDR": "localhost",
        "MQTT_PORT": 1883,
    }
    srv_name = "test"
    mt = MT(config, srv_name)
    app = Bluetooth()
    mt.Bind(app)
    mt.Run()


#! /usr/bin/env python3
# -*- coding: utf-8 -*-
from paho.mqtt import client as mqtt_client
import random
import time
from pdb import set_trace as stop
import json

class Err:
    SRV_NO_EXIST = 900 # 无此服务
    SRV_NO_RESP = 901 # 无回应
    SRV_REQ_INVALID = 902 # 无效 Req
    SRV_RESP_INVALID = 903 # 无效 Res

    DEV_NO_CONN = 1000 # 无法连接设备
    DEV_NO_RESP = 1001 # 无法得到回应
    DEV_REQ_INVALID = 1002 # 无效的 Request
    DEV_RESP_INVALID = 1003 # 无效的 Response

    BLE_RESP_INVALID = 11003 # 无效蓝牙回应
    BLE_NO_CHAR = 11100 # 无效蓝牙特征

class Alarmer():
    def __init__(self, broker: str = 'broker.emqx.io', port: int = 1883) -> None:
        self.broker = broker
        self.port = port
        self.client = None

    def connect(self, srv_id, dev_id, label, srv_type):
        def on_connect(client, userdata, flags, rc):
            if rc == 0:
                print("[Alarmer] Connected to MQTT Broker: %s:%s!" % (self.broker, self.port))
            else:
                print("[Alarmer] Failed to connect, return code %d\n" % rc)

        self.client = mqtt_client.Client(f'python-alarmer-{random.randint(0, 1000)}')
        self.client.on_connect = on_connect
        self.client.connect(self.broker, self.port)
        self.client.loop_start()
        self.topic = 'alarm/{0}/{1}'.format(srv_id, dev_id)
        self.alarm_msg = {
            "srv_id": srv_id,
            "dev_id": dev_id,
            "label": label,
            "srv_type": srv_type,
            "code_err": -1,
            "code_severity": -1,
            "time": 0,
            "content": ""
        }

    def alarm(self, code_err: int, code_severity: int, content: str):
        self.alarm_msg['code_err'] = code_err
        self.alarm_msg['code_severity'] = code_severity
        self.alarm_msg['content'] = content
        self.alarm_msg['time'] = time.time_ns()
        msg = json.dumps(self.alarm_msg, ensure_ascii=False)
        print("[ALARM]: " + msg)
        self.client.publish(self.topic, msg, 1, False)